import os
import sys
import sqlite3
import unittest
import lazysql
from lazysql import LazySql


class TestLazySql(unittest.TestCase):

    def setUp(self):
        if 'lazysql_test.db' in os.listdir():
            exit('lazysql_test.db already exists in current working directory.\nDelete the current instance or pick another directory to run test.')
        os.system('touch lazysql_test.db')

        with sqlite3.connect('lazysql_test.db') as conn:
            cur = conn.cursor()
            cur.execute("""CREATE TABLE test(
                        id INTEGER PRIMARY KEY,
                        _str STRING NOT NULL,
                        _int INTEGER NOT NULL,
                        _float FLOAT NOT NULL);""")
            conn.commit()
            for i in range(0, 10):
                cur.execute(
                    f"INSERT INTO test(_str, _int, _float) VALUES('test{i}', {i}, {i+0.5});")
            conn.commit()
        self.db = LazySql(sqlite3, 'lazysql_test.db')
        print('Setup complete, starting tests...')

    def tearDown(self):
        try:
            os.system('rm -rf lazysql_test.db')
            return True
        except Exception as e:
            print(e)
            return False

    def test_read(self):
        res = self.db.query("SELECT * FROM test WHERE _str='test1';")
        data = res[0]
        self.assertEqual(data['_str'], 'test1')
        self.assertEqual(data['_int'], 1)
        self.assertEqual(data['_float'], 1.5)

    def test_read_non_existant_data(self):
        res = self.db.query("SELECT * FROM test WHERE _str='invalidString';")
        self.assertEqual(res, [])

    def test_read_invalid_query(self):
        res = self.db.query(
            "SELECT invalid_table FROM test WHERE _str='test1';")
        self.assertEqual(res, False)

    def test_write(self):
        write_res = self.db.query(
            "INSERT INTO test(_str, _int, _float) VALUES('test11', 11, 11.5);", commit=True)
        read_res = self.db.query("SELECT * FROM test WHERE _str='test11';")
        data = read_res[0]
        self.assertEqual(write_res, True)
        self.assertEqual(data['_str'], 'test11')
        self.assertEqual(data['_int'], 11)
        self.assertEqual(data['_float'], 11.5)

    def test_async_read(self):
        queries = [{"query": f"SELECT * FROM test WHERE _str=?;", "data": [f"test{n}"]}
                   for n in range(1, 4)]
        res = self.db.async_query(queries)
        for i, q in enumerate(res):
            i += 1
            self.assertEqual(q[0]['_str'], f"test{i}")
            self.assertEqual(q[0]['_int'], i)
            self.assertEqual(q[0]['_float'], i+0.5)

    def test_async_read_with_non_existant_data(self):
        queries = [{"query": f"SELECT * FROM test WHERE _str=?;", "data": [f"test{n}"]}
                   for n in range(1, 4)]
        # Overwrite to make query2 error
        queries[1]['query'] = "SELECT * FROM test WHERE _str=?;"
        queries[1]['data'] = ['test11']
        query1, query2, query3 = self.db.async_query(queries)
        self.assertEqual(query1[0]['_str'], f"test1")
        self.assertEqual(query1[0]['_int'], 1)
        self.assertEqual(query1[0]['_float'], 1.5)
        self.assertEqual(query2, [])
        self.assertEqual(query3[0]['_str'], f"test3")
        self.assertEqual(query3[0]['_int'], 3)
        self.assertEqual(query3[0]['_float'], 3.5)

    def test_async_read_with_invalid_query(self):
        queries = [{"query": f"SELECT * FROM test WHERE _str=?;", "data": [f"test{n}"]}
                   for n in range(1, 4)]
        # Overwrite to make query2 error
        queries[1]['query'] = "SELECT * FROM invalid_table WHERE _str=?;"
        query1, query2, query3 = self.db.async_query(queries)
        self.assertEqual(query1[0]['_str'], f"test1")
        self.assertEqual(query1[0]['_int'], 1)
        self.assertEqual(query1[0]['_float'], 1.5)
        self.assertEqual(query2, False)
        self.assertEqual(query3[0]['_str'], f"test3")
        self.assertEqual(query3[0]['_int'], 3)
        self.assertEqual(query3[0]['_float'], 3.5)


if __name__ == '__main__':
    unittest.main()



import asyncio


class ConnectionNotFoundException(Exception):
    pass


class InvalidQueryException(Exception):
    pass


class LazySql:
    """
    connector: library object that calls .connect()
    uri: uri passed into .connect()
    max_conn: maximum number of simultaneous connections per query (for async)
    """

    def __init__(self, connector, uri, logger=None, max_conn=3):
        self.connector = connector  # DB connector i.e psycopg2, sqlite3
        self.uri = uri
        self.conn = None
        self.max_conn = max_conn
        if logger:
            self.logger = logger
        else:
            self.logger = None

    def query(self, query, data=None, commit=False):
        """
        Used for ephemeral operations
        Connection is closed after executing query
        """
        try:
            with self.connector.connect(self.uri) as conn:
                cur = conn.cursor()
                if not data:
                    cur.execute(query)
                else:
                    cur.execute(query, data)
                if not commit:
                    rows = cur.fetchall()
                    if not rows:
                        return rows
                    col_names = [c[0] for c in cur.description]
                    return [dict(zip(col_names, r)) for r in rows]
                else:
                    conn.commit()
                return True
        except Exception as e:
            logmsg = 'LazySql query Exception'
            print(f'{logmsg} {e}')
            if self.logger:
                self.logger.exception(f'{logmsg}')
            return False

    def async_query(self, query_list):
        """
        Queries are executed asynchoronously

        Be cautious when using with single connection databases like sqlite

        """
        async def __query(query, data=None, commit=False):
            try:
                with self.connector.connect(self.uri) as conn:
                    cur = conn.cursor()
                    if not data:
                        cur.execute({query})
                    else:
                        cur.execute(query, data)
                    if not commit:
                        rows = cur.fetchall()
                        if not rows:
                            return rows
                        col_names = [c[0] for c in cur.description]
                        return [dict(zip(col_names, r)) for r in rows]
                    else:
                        conn.commit()
                    return True
            except Exception as e:
                logmsg = 'LazySql query Exception'
                print(f'{logmsg} {e}')
                if self.logger:
                    self.logger.exception(f'{logmsg}')
                return False
        if not isinstance(query_list, list):
            raise TypeError(f"Invalid query, list required")
        coros = []
        res = []
        for q in query_list:
            if not isinstance(q, dict):
                raise TypeError(f"Invalid query {q}")
            if 'query' not in q:
                raise InvalidQueryException(
                    f"Required argument query not found")
            if 'data' not in q:
                q['data'] = None
            if 'commit' not in q:
                q['commit'] = False
            coros.append(
                __query(q['query'], data=q['data'], commit=q['commit']))
            if len(coros) >= self.max_conn:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                loop_res = loop.run_until_complete(asyncio.gather(*coros))
                res.extend(loop_res)
                loop.close()
                coros = []
        if coros:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop_res = loop.run_until_complete(asyncio.gather(*coros))
            res.extend(loop_res)
            loop.close()
        return res

    def batch(self, query, data=None, commit=False, close=False):
        """
        Used for write operations

        Db connection is manually handled
        commit will commit and close the connection
        close will close without committing
        """
        try:
            if all([close, commit]):
                raise ValueError('Choose commit or close')
            else:
                if not self.conn:
                    # Opens connection if not exist
                    self.conn = self.connector.connect(self.uri)
                    self.cur = self.conn.cursor()
                if query:
                    if not data:
                        self.cur.execute(query)
                    else:
                        self.cur.execute(query, data)
                if commit:
                    self.conn.commit()
                    close = True
                if close:
                    self.conn.close()
                    self.conn = None
                    self.cur = None
                return True
        except Exception as e:
            logmsg = 'LazySql batch Exception'
            print(f'{logmsg} {e}')
            if self.logger:
                self.logger.exception(f'{logmsg}')
            return False

    def commit(self):
        if self.conn:
            self.conn.commit()
            self.conn.close()
            return True
        raise ConnectionNotFoundException('DB connection not found')

    def close(self):
        if self.conn:
            self.conn.close()
            return True
        raise ConnectionNotFoundException('DB connection not found')

    def rollback(self):
        if self.conn:
            self.conn.rollback()
        raise ConnectionNotFoundException('DB connection not found')
